# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from typing import Generic

import sciline
import scipp as sc

from .types import (
    Analyzer,
    AnalyzerSpin,
    Down,
    FlipperEfficiency,
    HalfPolarizedCorrectedData,
    HalfPolarizedCorrection,
    NoAnalyzer,
    PolarizationCorrectedData,
    PolarizationCorrection,
    Polarizer,
    PolarizerSpin,
    PolarizingElement,
    PolarizingElementCorrection,
    ReducedSampleDataBySpinChannel,
    TransmissionFunction,
    Up,
)


@dataclass
class InverseFlipperMatrix(Generic[PolarizerSpin, PolarizingElement]):
    """Flipper matrix, combined with component flip for down component"""

    efficiency: FlipperEfficiency[PolarizingElement]
    swap: bool

    def from_left(
        self, up: sc.Variable, down: sc.Variable
    ) -> tuple[sc.Variable, sc.Variable]:
        """Apply inverse flipper matrix from the left (for analyzer)"""
        if self.swap:
            up, down = down, up
        f = 1 / self.efficiency.value
        if f == 1:
            return up, down
        return up, (1 - f) * up + f * down

    def from_right(
        self, up: sc.Variable, down: sc.Variable
    ) -> tuple[sc.Variable, sc.Variable]:
        """Apply inverse flipper matrix from the right (for polarizer)"""
        f = 1 / self.efficiency.value
        if f == 1:
            return (down, up) if self.swap else (up, down)
        if self.swap:
            return f * down, f * up
        else:
            return up + (1 - f) * down, down + (1 - f) * up


def make_spin_flipping_matrix_up(
    efficiency: FlipperEfficiency[PolarizingElement],
) -> InverseFlipperMatrix[Up, PolarizingElement]:
    return InverseFlipperMatrix[Up, PolarizingElement](
        efficiency=efficiency, swap=False
    )


def make_spin_flipping_matrix_down(
    efficiency: FlipperEfficiency[PolarizingElement],
) -> InverseFlipperMatrix[Down, PolarizingElement]:
    return InverseFlipperMatrix[Down, PolarizingElement](
        efficiency=efficiency, swap=True
    )


def compute_polarizing_element_correction(
    channel: ReducedSampleDataBySpinChannel[PolarizerSpin, AnalyzerSpin],
    transmission: TransmissionFunction[PolarizingElement],
) -> PolarizingElementCorrection[PolarizerSpin, AnalyzerSpin, PolarizingElement]:
    """
    Compute matrix coefficients for the correction of a polarizing element.

    The coefficients stem from the inverse of a symmetric matrix of the form
    [[Tplus, Tminus], [Tminus, Tplus]]. The inverse is given by a matrix
        mat = 1/denom * [[Tplus, -Tminus], [-Tminus, Tplus]],
    with
        denom = Tplus**2 - Tminus**2.
    As there are only two unique elements in the matrix, we return them as a dataclass
    with diagonal and off-diagonal elements.

    Parameters
    ----------
    channel :
        Data including wavelength (and time) for a given spin channel.
    transmission :
        Transmission function for the polarizing element.

    Returns
    -------
    :
        Correction matrix coefficients.
    """
    t_plus = transmission.apply(channel, 'plus')
    t_minus = transmission.apply(channel, 'minus')
    t_minus *= -1
    denom = t_plus**2 - t_minus**2
    sc.reciprocal(denom, out=denom)
    t_plus *= denom
    t_minus *= denom
    return PolarizingElementCorrection[PolarizerSpin, AnalyzerSpin, PolarizingElement](
        diag=t_plus, off_diag=t_minus
    )


def compute_polarization_correction(
    *,
    analyzer: PolarizingElementCorrection[PolarizerSpin, AnalyzerSpin, Analyzer],
    polarizer: PolarizingElementCorrection[PolarizerSpin, AnalyzerSpin, Polarizer],
    analyzer_flipper: InverseFlipperMatrix[AnalyzerSpin, Analyzer],
    polarizer_flipper: InverseFlipperMatrix[PolarizerSpin, Polarizer],
) -> PolarizationCorrection[PolarizerSpin, AnalyzerSpin]:
    """
    Compute columns of combined correction coefficients for polarizer and analyzer.

    This is effectively a column resulting from a sparse matrix-matrix product.

    Parameters
    ----------
    analyzer :
        Correction coefficients for the analyzer.
    polarizer :
        Correction coefficients for the polarizer.
    analyzer_flipper :
        Flipper matrix for the analyzer.
    polarizer_flipper :
        Flipper matrix for the polarizer.

    Returns
    -------
    :
        Combined correction coefficients.
    """
    a_up, a_down = analyzer_flipper.from_left(analyzer.diag, analyzer.off_diag)
    p_up, p_down = polarizer_flipper.from_right(polarizer.diag, polarizer.off_diag)
    return PolarizationCorrection[PolarizerSpin, AnalyzerSpin](
        upup=p_up * a_up,
        updown=p_up * a_down,
        downup=p_down * a_up,
        downdown=p_down * a_down,
    )


def compute_polarization_corrected_data(
    channel: ReducedSampleDataBySpinChannel[PolarizerSpin, AnalyzerSpin],
    polarization_correction: PolarizationCorrection[PolarizerSpin, AnalyzerSpin],
) -> PolarizationCorrectedData[PolarizerSpin, AnalyzerSpin]:
    # TODO Would like to use inplace ops, but modifying input is dodgy. Maybe combine
    # into a single function?
    return PolarizationCorrectedData(
        upup=channel * polarization_correction.upup,
        updown=channel * polarization_correction.updown,
        downup=channel * polarization_correction.downup,
        downdown=channel * polarization_correction.downdown,
    )


def compute_half_polarized_correction(
    *,
    polarizer: PolarizingElementCorrection[PolarizerSpin, NoAnalyzer, Polarizer],
    polarizer_flipper: InverseFlipperMatrix[PolarizerSpin, Polarizer],
) -> HalfPolarizedCorrection[PolarizerSpin]:
    p_up, p_down = polarizer_flipper.from_right(polarizer.diag, polarizer.off_diag)
    return HalfPolarizedCorrection[PolarizerSpin](up=p_up, down=p_down)


def compute_half_polarized_corrected_data(
    channel: ReducedSampleDataBySpinChannel[PolarizerSpin, NoAnalyzer],
    polarization_correction: HalfPolarizedCorrection[PolarizerSpin],
) -> HalfPolarizedCorrectedData[PolarizerSpin]:
    return HalfPolarizedCorrectedData(
        up=channel * polarization_correction.up,
        down=channel * polarization_correction.down,
    )


def CorrectionWorkflow(half_polarized: bool = False) -> sciline.Pipeline:
    workflow = sciline.Pipeline(
        (
            make_spin_flipping_matrix_up,
            make_spin_flipping_matrix_down,
            compute_polarizing_element_correction,
        )
    )
    if half_polarized:
        workflow.insert(compute_half_polarized_correction)
        workflow.insert(compute_half_polarized_corrected_data)
    else:
        workflow.insert(compute_polarization_correction)
        workflow.insert(compute_polarization_corrected_data)
    # If there is no flipper, setting an efficiency of 1.0 is equivalent to not using
    # a flipper.
    workflow[FlipperEfficiency[PolarizingElement]] = FlipperEfficiency[
        PolarizingElement
    ](value=1.0)
    return workflow


def PolarizationAnalysisWorkflow(
    *,
    polarizer_workflow: sciline.Pipeline,
    analyzer_workflow: sciline.Pipeline,
) -> sciline.Pipeline:
    """
    Create a polarization analysis workflow.

    Parameters
    ----------
    polarizer_workflow :
        Workflow for the polarizer, e.g., a He3CellWorkflow or SupermirrorWorkflow.
    analyzer_workflow :
        Workflow for the analyzer, e.g., a He3CellWorkflow or SupermirrorWorkflow.

    Returns
    -------
    :
        Full workflow for polarization analysis.
    """

    workflow = CorrectionWorkflow()
    workflow[TransmissionFunction[Polarizer]] = polarizer_workflow[
        TransmissionFunction[Polarizer]
    ]
    workflow[TransmissionFunction[Analyzer]] = analyzer_workflow[
        TransmissionFunction[Analyzer]
    ]
    return workflow


def HalfPolarizedWorkflow(
    *,
    polarizer_workflow: sciline.Pipeline,
) -> sciline.Pipeline:
    """
    Create a half-polarized workflow, i.e, with a polarizer but no analyzer.

    Parameters
    ----------
    polarizer_workflow :
        Workflow for the polarizer, e.g., a He3CellWorkflow or SupermirrorWorkflow.

    Returns
    -------
    :
        Half-polarized workflow.
    """
    workflow = CorrectionWorkflow(half_polarized=True)
    workflow[TransmissionFunction[Polarizer]] = polarizer_workflow[
        TransmissionFunction[Polarizer]
    ]
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from typing import Generic, Sequence

import ess.isissans as isis
import mantid.api as _mantid_api
import sciline as sl
import scipp as sc
from ess.isissans.data import LoadedFileContents
from ess.isissans.mantidio import DataWorkspace, Period
from ess.sans.types import (
    Filename,
    Incident,
    MonitorType,
    NeXusMonitorName,
    RawMonitor,
    RunType,
    SampleRun,
    Transmission,
    TransmissionRun,
    UncertaintyBroadcastMode,
)
from mantid import simpleapi as _mantid_simpleapi

# In this case the "sample" is the analyzer cell, of which we want to measure
# the transmission fraction.
sample_run_type = TransmissionRun[SampleRun]


def load_histogrammed_run(
    filename: Filename[sample_run_type], period: Period
) -> DataWorkspace[sample_run_type]:
    """Load a non-event-data ISIS file"""
    loaded = _mantid_simpleapi.Load(Filename=str(filename), StoreInADS=False)
    if isinstance(loaded, _mantid_api.Workspace):
        # A single workspace
        data_ws = loaded
        if isinstance(data_ws, _mantid_api.WorkspaceGroup):
            if period is None:
                raise ValueError(
                    f'Needs {Period} to be set to know what '
                    'section of the event data to load'
                )
            data_ws = data_ws.getItem(period)
    else:
        # Separate data and monitor workspaces
        data_ws = loaded.OutputWorkspace
        if isinstance(data_ws, _mantid_api.WorkspaceGroup):
            if period is None:
                raise ValueError(
                    f'Needs {Period} to be set to know what '
                    'section of the event data to load'
                )
            data_ws = data_ws.getItem(period)
            data_ws.setMonitorWorkspace(loaded.MonitorWorkspace.getItem(period))
        else:
            data_ws.setMonitorWorkspace(loaded.MonitorWorkspace)
    return DataWorkspace[sample_run_type](data_ws)


def _get_time(dg: sc.DataGroup) -> sc.Variable:
    start = sc.datetime(dg['run_start'].value)
    end = sc.datetime(dg['run_end'].value)
    delta = end - start
    return start + delta // 2


def _get_time_dependent_monitor(*monitors: sc.DataArray) -> sc.DataArray:
    monitors = sc.concat(monitors, 'time')
    datetime = monitors.coords['datetime']
    monitors.coords['time'] = datetime - datetime.min()
    del monitors.coords['spectrum']
    del monitors.coords['detector_id']
    return monitors


@dataclass
class MonitorSpectrumNumber(Generic[MonitorType]):
    value: int


def get_monitor_data(
    dg: LoadedFileContents[RunType], nexus_name: NeXusMonitorName[MonitorType]
) -> RawMonitor[RunType, MonitorType]:
    """
    Same as :py:func:`ess.isissans.get_monitor_data` but dropping variances.

    Dropping variances is a workaround required since ESSsans does not handle
    variance broadcasting when combining monitors. In our case some of the monitors
    are time-dependent, so this is required for now.
    """
    # See https://github.com/scipp/sciline/issues/52 why copy needed
    mon = dg['monitors'][nexus_name]['data'].copy()
    return RawMonitor[RunType, MonitorType](sc.values(mon))


def get_monitor_data_from_transmission_run(
    dg: LoadedFileContents[TransmissionRun[RunType]],
    spectrum_number: MonitorSpectrumNumber[MonitorType],
) -> RawMonitor[TransmissionRun[RunType], MonitorType]:
    """
    Extract incident or transmission monitor from ZOOM direct-beam run

    The files in this case do not contain detector data, only monitor data. Mantid
    stores this as a Workspace2D, where each spectrum corresponds to a monitor.
    """
    # Note we index with a scipp.Variable, i.e., by the spectrum number used at ISIS
    monitor = dg['data']['spectrum', sc.index(spectrum_number.value)].copy()
    monitor.coords['datetime'] = _get_time(dg)
    return monitor


def ZoomTransmissionFractionWorkflow(runs: Sequence[str]) -> sl.Pipeline:
    """
    Workflow computing time-dependent SANS transmission fraction from ZOOM data.

    The time-dependence is obtained by using a sequence of runs.

    .. code-block:: python

        workflow = ZoomTransmissionFractionWorkflow(cell_runs)

    Note that in this case the "sample" (of which the transmission is to be computed)
    is the He3 analyzer cell.

    Parameters
    ----------
    runs:
        List of filenames of the runs to use for the transmission fraction.
    """
    workflow = isis.zoom.ZoomWorkflow()
    workflow.insert(get_monitor_data)
    workflow.insert(get_monitor_data_from_transmission_run)
    workflow.insert(load_histogrammed_run)

    mapped = workflow.map({Filename[TransmissionRun[SampleRun]]: runs})
    for mon_type in (Incident, Transmission):
        workflow[RawMonitor[TransmissionRun[SampleRun], mon_type]] = mapped[
            RawMonitor[TransmissionRun[SampleRun], mon_type]
        ].reduce(func=_get_time_dependent_monitor)

    # We are dealing with two different types of files, and monitors are identified
    # differently in each case, so there is some duplication here.
    workflow[MonitorSpectrumNumber[Incident]] = MonitorSpectrumNumber[Incident](3)
    workflow[MonitorSpectrumNumber[Transmission]] = MonitorSpectrumNumber[Transmission](
        5
    )
    workflow[NeXusMonitorName[Incident]] = NeXusMonitorName[Incident]('monitor3')
    workflow[NeXusMonitorName[Transmission]] = NeXusMonitorName[Transmission](
        'monitor5'
    )
    workflow[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound

    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

# flake8: noqa
import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib


from .correction import (
    CorrectionWorkflow,
    HalfPolarizedWorkflow,
    PolarizationAnalysisWorkflow,
)
from .he3 import (
    Depolarized,
    DirectBeamBackgroundQRange,
    DirectBeamNoCell,
    DirectBeamQRange,
    He3CellLength,
    He3CellPressure,
    He3CellWorkflow,
    He3DirectBeam,
    He3FillingTime,
    He3Opacity0,
    He3OpacityFunction,
    He3PolarizationFunction,
    He3TransmissionEmptyGlass,
    He3TransmissionFunction,
    Polarized,
)
from .supermirror import (
    SecondDegreePolynomialEfficiency,
    SupermirrorEfficiencyFunction,
    SupermirrorWorkflow,
)
from .types import (
    Analyzer,
    Down,
    HalfPolarizedCorrectedData,
    NoAnalyzer,
    PolarizationCorrectedData,
    Polarizer,
    PolarizingElement,
    ReducedSampleDataBySpinChannel,
    TransmissionFunction,
    Up,
)

__all__ = [
    "Analyzer",
    "CorrectionWorkflow",
    "Depolarized",
    "DirectBeamBackgroundQRange",
    "DirectBeamNoCell",
    "DirectBeamQRange",
    "Down",
    "HalfPolarizedWorkflow",
    "HalfPolarizedCorrectedData",
    "He3CellLength",
    "He3CellPressure",
    "He3CellWorkflow",
    "He3DirectBeam",
    "He3FillingTime",
    "He3Opacity0",
    "He3OpacityFunction",
    "He3PolarizationFunction",
    "He3TransmissionEmptyGlass",
    "He3TransmissionFunction",
    "PolarizationAnalysisWorkflow",
    "PolarizationCorrectedData",
    "Polarizer",
    "PolarizingElement",
    "ReducedSampleDataBySpinChannel",
    "SecondDegreePolynomialEfficiency",
    "SupermirrorEfficiencyFunction",
    "SupermirrorWorkflow",
    "TransmissionFunction",
    "NoAnalyzer",
    "Up",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic

import sciline
import scipp as sc

from .types import PlusMinus, PolarizingElement, TransmissionFunction


class SupermirrorEfficiencyFunction(Generic[PolarizingElement], ABC):
    """Base class for supermirror efficiency functions"""

    @abstractmethod
    def __call__(self, *, wavelength: sc.Variable) -> sc.DataArray:
        """Return the efficiency of a supermirror for a given wavelength"""


@dataclass
class SecondDegreePolynomialEfficiency(
    SupermirrorEfficiencyFunction[PolarizingElement]
):
    """
    Efficiency of a supermirror as a second-degree polynomial

    The efficiency is given by a * wavelength^2 + b * wavelength + c

    Parameters
    ----------
    a:
        Coefficient of the quadratic term, with unit of 1/angstrom^2
    b:
        Coefficient of the linear term, with unit of 1/angstrom
    c:
        Constant term, dimensionless
    """

    a: sc.Variable
    b: sc.Variable
    c: sc.Variable

    def __call__(self, *, wavelength: sc.Variable) -> sc.DataArray:
        """Return the efficiency of a supermirror for a given wavelength"""
        return (
            (self.a * wavelength**2).to(unit='', copy=False)
            + (self.b * wavelength).to(unit='', copy=False)
            + self.c.to(unit='', copy=False)
        )


@dataclass
class SupermirrorTransmissionFunction(TransmissionFunction[PolarizingElement]):
    """Wavelength-dependent transmission of a supermirror"""

    efficiency_function: SupermirrorEfficiencyFunction

    def __call__(
        self, *, wavelength: sc.Variable, plus_minus: PlusMinus
    ) -> sc.DataArray:
        """Return the transmission fraction for a given wavelength"""
        efficiency = self.efficiency_function(wavelength=wavelength)
        if plus_minus == 'plus':
            return 0.5 * (1 + efficiency)
        else:
            return 0.5 * (1 - efficiency)

    def apply(self, data: sc.DataArray, plus_minus: PlusMinus) -> sc.DataArray:
        """Apply the transmission function to a data array"""
        return self(wavelength=data.coords['wavelength'], plus_minus=plus_minus)


def get_supermirror_transmission_function(
    efficiency_function: SupermirrorEfficiencyFunction[PolarizingElement],
) -> TransmissionFunction[PolarizingElement]:
    return SupermirrorTransmissionFunction[PolarizingElement](
        efficiency_function=efficiency_function
    )


def SupermirrorWorkflow() -> sciline.Pipeline:
    """
    Workflow for computing transmission functions for supermirror polarizing elements.
    """
    return sciline.Pipeline((get_supermirror_transmission_function,))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc

import ess.polarization as pol


def test_SecondDegreePolynomialEfficiency_raises_if_units_incompatible():
    wav = sc.scalar(1.0, unit='m')
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff = pol.SecondDegreePolynomialEfficiency(
            a=sc.scalar(1.0, unit='1/angstrom'),
            b=sc.scalar(1.0, unit='1/angstrom'),
            c=sc.scalar(1.0),
        )
        eff(wavelength=wav)
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff = pol.SecondDegreePolynomialEfficiency(
            a=sc.scalar(1.0, unit='1/angstrom**2'),
            b=sc.scalar(1.0, unit='1/angstrom**2'),
            c=sc.scalar(1.0),
        )
        eff(wavelength=wav)
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff = pol.SecondDegreePolynomialEfficiency(
            a=sc.scalar(1.0, unit='1/angstrom**2'),
            b=sc.scalar(1.0, unit='1/angstrom'),
            c=sc.scalar(1.0, unit='1/angstrom'),
        )
        eff(wavelength=wav)
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff = pol.SecondDegreePolynomialEfficiency(
            a=sc.scalar(1.0, unit='1/angstrom**2'),
            b=sc.scalar(1.0, unit='1/angstrom'),
            c=sc.scalar(1.0),
        )
        eff(wavelength=wav / sc.scalar(1.0, unit='s'))


def test_SecondDegreePolynomialEfficiency_produces_correct_values():
    a = sc.scalar(1.0, unit='1/angstrom**2')
    b = sc.scalar(2.0, unit='1/angstrom')
    c = sc.scalar(3.0)
    f = pol.SecondDegreePolynomialEfficiency(a=a, b=b, c=c)
    assert f(wavelength=sc.scalar(0.0, unit='angstrom')) == 3.0
    assert f(wavelength=sc.scalar(1.0, unit='angstrom')) == 6.0
    assert f(wavelength=sc.scalar(2.0, unit='angstrom')) == 11.0


def test_SecondDegreePolynomialEfficiency_converts_units():
    a = sc.scalar(1.0, unit='1/angstrom**2')
    b = sc.scalar(20.0, unit='1/nm')
    c = sc.scalar(3.0)
    f = pol.SecondDegreePolynomialEfficiency(a=a, b=b, c=c)
    assert f(wavelength=sc.scalar(0.0, unit='angstrom')) == 3.0
    assert f(wavelength=sc.scalar(1.0, unit='angstrom')) == 6.0
    assert f(wavelength=sc.scalar(2.0, unit='angstrom')) == 11.0
    assert f(wavelength=sc.scalar(0.0, unit='nm')) == 3.0
    assert f(wavelength=sc.scalar(0.1, unit='nm')) == 6.0
    assert f(wavelength=sc.scalar(0.2, unit='nm')) == 11.0



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import scipp as sc

from ess.polarization import he3


def test_incoming_unpolarized_reproduces_input_params_within_errors() -> None:
    time = sc.linspace('time', 0.0, 1000000.0, num=1000, unit='s')
    wavelength = sc.linspace('wavelength', 0.5, 5.0, num=40, unit='angstrom')
    C = sc.scalar(1.3)
    T1 = sc.scalar(123456.0, unit='s')
    opacity0 = sc.scalar(0.88, unit='1/angstrom')
    polarization_function = he3.He3PolarizationFunction(C=C, T1=T1)
    opacity_function = he3.He3OpacityFunction(opacity0)
    transmission_empty_glass = sc.scalar(0.9)
    opacity = opacity_function(wavelength)
    polarization = polarization_function(time)
    transmission = he3.transmission_incoming_unpolarized(
        transmission_empty_glass=transmission_empty_glass,
        opacity=opacity,
        polarization=polarization,
    )

    result = he3.get_he3_transmission_incoming_unpolarized_from_fit_to_direct_beam(
        transmission_fraction=transmission,
        opacity_function=opacity_function,
        transmission_empty_glass=transmission_empty_glass,
    )
    polarization_function = result.polarization_function
    assert isinstance(polarization_function, he3.He3PolarizationFunction)

    # No noise, very close or exact match.
    assert sc.isclose(polarization_function.C, C)
    assert sc.isclose(polarization_function.T1, T1)

    rng = np.random.default_rng(seed=1234)
    transmission_noisy = transmission.copy()
    transmission_noisy.values += rng.normal(0.0, 0.01, transmission_noisy.shape)

    result = he3.get_he3_transmission_incoming_unpolarized_from_fit_to_direct_beam(
        transmission_fraction=transmission_noisy,
        opacity_function=opacity_function,
        transmission_empty_glass=transmission_empty_glass,
    )
    polarization_function = result.polarization_function

    # With noise, within 1% of the input values.
    assert sc.isclose(polarization_function.C, C, rtol=sc.scalar(1e-2))
    assert sc.isclose(polarization_function.T1, T1, rtol=sc.scalar(1e-2))


def test_incoming_polarized_reproduces_input_params_within_errors() -> None:
    time = sc.linspace('time', 0.0, 1000000.0, num=1000, unit='s')
    wavelength = sc.linspace('wavelength', 0.5, 5.0, num=40, unit='angstrom')
    C = sc.scalar(1.3)
    T1 = sc.scalar(123456.0, unit='s')
    opacity0 = sc.scalar(0.88, unit='1/angstrom')
    polarization_function = he3.He3PolarizationFunction(C=C, T1=T1)
    opacity_function = he3.He3OpacityFunction(opacity0)
    transmission_empty_glass = sc.scalar(0.9)
    transmission_function = he3.He3TransmissionFunction(
        transmission_empty_glass=transmission_empty_glass,
        opacity_function=opacity_function,
        polarization_function=polarization_function,
    )
    # State switch at 456th time point, cut further below into 4 channels total.
    plus = transmission_function(
        time=time[:456], wavelength=wavelength, plus_minus='plus'
    )
    minus = transmission_function(
        time=time[456:], wavelength=wavelength, plus_minus='minus'
    )

    result = he3.get_he3_transmission_incoming_polarized_from_fit_to_direct_beam(
        plus=plus,
        minus=minus,
        opacity_function=opacity_function,
        transmission_empty_glass=transmission_empty_glass,
    )
    polarization_function = result.polarization_function
    assert isinstance(polarization_function, he3.He3PolarizationFunction)

    # No noise, very close or exact match.
    assert sc.isclose(polarization_function.C, C)
    assert sc.isclose(polarization_function.T1, T1)

    rng = np.random.default_rng(seed=1234)
    plus_noisy = plus.copy()
    minus_noisy = minus.copy()
    plus_noisy.values += rng.normal(0.0, 0.01, plus_noisy.shape)
    minus_noisy.values += rng.normal(0.0, 0.01, minus_noisy.shape)

    result = he3.get_he3_transmission_incoming_polarized_from_fit_to_direct_beam(
        plus=plus_noisy,
        minus=minus_noisy,
        opacity_function=opacity_function,
        transmission_empty_glass=transmission_empty_glass,
    )
    polarization_function = result.polarization_function

    # With noise, within 1% of the input values.
    assert sc.isclose(polarization_function.C, C, rtol=sc.scalar(1e-2))
    assert sc.isclose(polarization_function.T1, T1, rtol=sc.scalar(1e-2))


def test_incoming_polarized_raises_if_plus_minus_coord_is_bad() -> None:
    time = sc.linspace('time', 0.0, 1000000.0, num=1000, unit='s')
    wavelength = sc.linspace('wavelength', 0.5, 5.0, num=40, unit='angstrom')
    C = sc.scalar(1.3)
    T1 = sc.scalar(123456.0, unit='s')
    opacity0 = sc.scalar(0.88, unit='1/angstrom')
    polarization_function = he3.He3PolarizationFunction(C=C, T1=T1)
    opacity_function = he3.He3OpacityFunction(opacity0)
    transmission_empty_glass = sc.scalar(0.9)
    transmission_function = he3.He3TransmissionFunction(
        transmission_empty_glass=transmission_empty_glass,
        opacity_function=opacity_function,
        polarization_function=polarization_function,
    )
    # State switch at 456th time point, cut further below into 4 channels total.
    plus = transmission_function(
        time=time[:456], wavelength=wavelength, plus_minus='plus'
    )
    minus = transmission_function(
        time=time[456:], wavelength=wavelength, plus_minus='minus'
    )

    with pytest.raises(ValueError, match='plus-minus coordinate of plus channel'):
        he3.get_he3_transmission_incoming_polarized_from_fit_to_direct_beam(
            plus=plus.assign_coords({'plus_minus': sc.scalar(-1)}),
            minus=minus,
            opacity_function=opacity_function,
            transmission_empty_glass=transmission_empty_glass,
        )
    with pytest.raises(ValueError, match='plus-minus coordinate of minus channel'):
        he3.get_he3_transmission_incoming_polarized_from_fit_to_direct_beam(
            plus=plus,
            minus=minus.assign_coords({'plus_minus': sc.scalar(1)}),
            opacity_function=opacity_function,
            transmission_empty_glass=transmission_empty_glass,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from typing import Generic, NewType, TypeVar

import sciline as sl
import scipp as sc

from .types import (
    Analyzer,
    AnalyzerSpin,
    Down,
    PlusMinus,
    PolarizerSpin,
    PolarizingElement,
    TransmissionFunction,
    Up,
)
from .uncertainty import broadcast_with_upper_bound_variances

Depolarized = NewType('Depolarized', int)
Polarized = NewType('Polarized', int)
"""Polarized either up or down, don't care."""
PolarizationState = TypeVar('PolarizationState', Polarized, Depolarized)


DirectBeamNoCell = NewType('DirectBeamNoCell', sc.DataArray)
"""Direct beam without cells and sample as a function of wavelength."""


class He3CellPressure(sl.Scope[PolarizingElement, sc.Variable], sc.Variable):
    """Pressure for a given cell."""


class He3CellLength(sl.Scope[PolarizingElement, sc.Variable], sc.Variable):
    """Length for a given cell."""


class He3CellTemperature(sl.Scope[PolarizingElement, sc.Variable], sc.Variable):
    """Temperature for a given cell."""


class He3FillingTime(sl.Scope[PolarizingElement, sc.Variable], sc.Variable):
    """Filling wall-clock time for a given cell."""


class He3CellTransmissionFractionIncomingUnpolarized(
    sl.ScopeTwoParams[PolarizingElement, PolarizationState, sc.DataArray], sc.DataArray
):
    """Transmission fraction for a given cell"""


class He3AnalyzerTransmissionFractionIncomingPolarized(
    sl.ScopeTwoParams[PolarizerSpin, AnalyzerSpin, sc.DataArray], sc.DataArray
):
    """Transmission fraction of the analyzer with polarized incoming beam"""


He3AnalyzerTransmissionFractionParallel = NewType(
    'He3AnalyzerTransmissionFractionParallel', sc.DataArray
)
"""Transmission fraction of analyzer with parallel polarized incoming beam"""


He3AnalyzerTransmissionFractionAntiParallel = NewType(
    'He3AnalyzerTransmissionFractionAntiParallel', sc.DataArray
)
"""Transmission fraction of analyzer with anti-parallel polarized incoming beam"""


class He3TransmissionEmptyGlass(sl.Scope[PolarizingElement, sc.Variable], sc.Variable):
    """Transmission of the empty glass for a given cell."""


class He3DirectBeam(
    sl.ScopeTwoParams[PolarizingElement, PolarizationState, sc.DataArray], sc.DataArray
):
    """
    Direct beam data for a given cell and spin state as a function of wavelength.
    """


class He3Opacity0(sl.Scope[PolarizingElement, sc.Variable], sc.Variable):
    """Opacity at 1 Angstrom for a given cell."""


class He3OpacityFunction(Generic[PolarizingElement]):
    """Wavelength-dependent opacity function for a given cell."""

    def __init__(self, opacity0: sc.Variable):
        self._opacity0 = opacity0.to(unit='1/Angstrom')

    @property
    def opacity0(self) -> sc.Variable:
        return self._opacity0

    def __call__(self, wavelength: sc.Variable) -> sc.Variable:
        scale = broadcast_with_upper_bound_variances(
            self.opacity0, sizes=wavelength.sizes
        )
        return sc.DataArray(
            (scale * wavelength).to(unit='', copy=False),
            coords={'wavelength': wavelength},
        )


def he3_opacity_from_cell_params(
    pressure: He3CellPressure[PolarizingElement],
    length: He3CellLength[PolarizingElement],
    temperature: He3CellTemperature[PolarizingElement],
) -> He3Opacity0[PolarizingElement]:
    """Opacity 0 for a given cell, estimated from pressure and cell length."""
    from scipp.constants import Boltzmann as k_B

    he3_neutron_absorption_cross_section_at_1_angstrom = 2966.0e-24 * sc.Unit(
        'cm^2/Angstrom'
    )
    # Try to convert to Kelvin, since we get a more cryptic error message later if we
    # do not, in case the user passes in a temperature in degree Celsius.
    temperature = temperature.to(unit='K')

    opacity0 = (
        he3_neutron_absorption_cross_section_at_1_angstrom
        / (k_B * temperature)
        * pressure
        * length
    )
    return He3Opacity0[PolarizingElement](opacity0)


def he3_opacity_function_from_cell_opacity(
    opacity0: He3Opacity0[PolarizingElement],
) -> He3OpacityFunction[PolarizingElement]:
    """
    Opacity function for a given cell, based on pressure and cell length.

    Note that this can alternatively be defined via neutron beam data, see
    :py:func:`he3_opacity_function_from_beam_data`.
    """
    return He3OpacityFunction[PolarizingElement](opacity0)


def _with_midpoints(data: sc.DataArray, dim: str) -> sc.DataArray:
    if data.coords.is_edges(dim):
        return data.assign_coords({dim: sc.midpoints(data.coords[dim])})
    return data


def he3_opacity_function_from_beam_data(
    transmission_empty_glass: He3TransmissionEmptyGlass[PolarizingElement],
    transmission_fraction: He3CellTransmissionFractionIncomingUnpolarized[
        PolarizingElement, Depolarized
    ],
    opacity0_initial_guess: He3Opacity0[PolarizingElement],
) -> He3OpacityFunction[PolarizingElement]:
    """
    Opacity function for a given cell, based on direct beam data.

    Note that this can alternatively be defined via cell parameters, see
    :py:func:`he3_opacity_function_from_cell_opacity`. The cell opacity is used as an
    initial guess for the fit.
    """

    # TODO Fit the exponent, since too much weight on low wavelengths?
    def intensity(wavelength: sc.Variable, opacity0: sc.Variable) -> sc.Variable:
        opacity = He3OpacityFunction[PolarizingElement](opacity0)
        return transmission_empty_glass * sc.exp(-opacity(wavelength))

    popt, _ = sc.curve_fit(
        ['wavelength'],
        intensity,
        _with_midpoints(transmission_fraction, 'wavelength'),
        p0={'opacity0': opacity0_initial_guess},
    )
    return He3OpacityFunction[PolarizingElement](sc.values(popt['opacity0']).data)


class He3PolarizationFunction(Generic[PolarizingElement]):
    """Time-dependent polarization function for a given cell."""

    def __init__(self, C: sc.Variable, T1: sc.Variable):
        self._C = C
        self._T1 = T1

    @property
    def C(self) -> sc.Variable:
        return self._C

    @property
    def T1(self) -> sc.Variable:
        return self._T1

    def __call__(self, time: sc.Variable) -> sc.Variable:
        return sc.DataArray(self.C * sc.exp(-time / self.T1), coords={'time': time})


@dataclass
class He3TransmissionFunction(TransmissionFunction[PolarizingElement]):
    """Wavelength- and time-dependent transmission for a given cell."""

    opacity_function: He3OpacityFunction[PolarizingElement]
    polarization_function: He3PolarizationFunction[PolarizingElement]
    transmission_empty_glass: He3TransmissionEmptyGlass[PolarizingElement]

    def __call__(
        self, *, time: sc.Variable, wavelength: sc.Variable, plus_minus: PlusMinus
    ) -> sc.Variable:
        opacity = self.opacity_function(wavelength)
        polarization = self.polarization_function(time)
        if plus_minus == 'plus':
            polarization *= -1.0
        elif isinstance(plus_minus, sc.Variable):
            polarization *= -plus_minus
        return self.transmission_empty_glass * sc.exp(-opacity * (1.0 + polarization))

    def apply(self, data: sc.DataArray, plus_minus: PlusMinus) -> sc.DataArray:
        return self(
            time=data.coords['time'],
            wavelength=data.coords['wavelength'],
            plus_minus=plus_minus,
        )


def transmission_incoming_unpolarized(
    *,
    transmission_empty_glass: sc.Variable,
    opacity: sc.Variable,
    polarization: sc.Variable,
) -> sc.Variable:
    return transmission_empty_glass * sc.exp(-opacity) * sc.cosh(opacity * polarization)


def compute_transmission_fraction_from_direct_beam(
    direct_beam_no_cell: DirectBeamNoCell,
    direct_beam_polarized: He3DirectBeam[PolarizingElement, PolarizationState],
) -> He3CellTransmissionFractionIncomingUnpolarized[
    PolarizingElement, PolarizationState
]:
    """
    Compute the transmission fraction for a given cell and polarization state.

    This is defined as the ratio of the direct beam with the cell to the direct beam
    without the cell. The result is a function of wavelength and time.

    Note that this is possible only if the main detector is used to measure direct
    beam data. If direct beam is computed from monitors then, e.g., the SANS
    transmission fraction (as computed by a regular SANS workflow) should be used
    directly. Note that the regular SANS workflow also normalized to an empty beam
    run, make sure to not perform the division twice.
    """
    return He3CellTransmissionFractionIncomingUnpolarized[
        PolarizingElement, PolarizationState
    ](direct_beam_polarized / direct_beam_no_cell)


def get_he3_transmission_incoming_unpolarized_from_fit_to_direct_beam(
    transmission_fraction: He3CellTransmissionFractionIncomingUnpolarized[
        PolarizingElement, Polarized
    ],
    opacity_function: He3OpacityFunction[PolarizingElement],
    transmission_empty_glass: He3TransmissionEmptyGlass[PolarizingElement],
) -> TransmissionFunction[PolarizingElement]:
    """
    Transmission function for a given cell, with unpolarized incoming beam.

    This is composed from the opacity-function and the polarization-function.
    The implementation fits a time- and wavelength-dependent equation and returns
    the fitted T(t, lambda).

    DB_pol/DB = T_E * cosh(O(lambda)*P(t))*exp(-O(lambda))
    """

    def expected_transmission(
        wavelength: sc.Variable, time: sc.Variable, C: sc.Variable, T1: sc.Variable
    ) -> sc.Variable:
        polarization_function = He3PolarizationFunction[PolarizingElement](C=C, T1=T1)
        opacity = opacity_function(wavelength)
        polarization = polarization_function(time)
        return transmission_incoming_unpolarized(
            transmission_empty_glass=transmission_empty_glass,
            opacity=opacity,
            polarization=polarization,
        )

    popt, _ = sc.curve_fit(
        ['wavelength', 'time'],
        expected_transmission,
        _with_midpoints(transmission_fraction, 'wavelength'),
        p0={'C': sc.scalar(0.8, unit=''), 'T1': sc.scalar(400000.0, unit='s')},
    )
    # TODO Consider including variances from fit
    polarization_function = He3PolarizationFunction[PolarizingElement](
        C=sc.values(popt['C']).data, T1=sc.values(popt['T1']).data
    )
    return He3TransmissionFunction[PolarizingElement](
        opacity_function=opacity_function,
        polarization_function=polarization_function,
        transmission_empty_glass=transmission_empty_glass,
    )


def transmission_fraction_analyzer_parallel(
    upup: He3AnalyzerTransmissionFractionIncomingPolarized[Up, Up],
    downdown: He3AnalyzerTransmissionFractionIncomingPolarized[Down, Down],
) -> He3AnalyzerTransmissionFractionParallel:
    """
    Analyzer Transmission fraction with polarization parallel to incoming beam.

    It may not always we desirable to use both up-up and down-down transmission,
    fractions. If that is the case, set He3AnalyzerTransmissionFractionPlus directly
    instead of using this helper.
    """
    # The transmission fractions do not share a common time coordinate. Therefore,
    # we cannot concat along a third dimension for the fit, but concat along time,
    # with an additional coordinate for whether polarizations are parallel or
    # antiparallel.
    return sc.concat([upup, downdown], 'time').assign_coords(plus_minus=sc.scalar(1))


def transmission_fraction_analyzer_antiparallel(
    updown: He3AnalyzerTransmissionFractionIncomingPolarized[Up, Down],
    downup: He3AnalyzerTransmissionFractionIncomingPolarized[Down, Up],
) -> He3AnalyzerTransmissionFractionAntiParallel:
    """
    Analyzer transmission fraction with polarization anti-parallel to incoming beam.

    It may not always we desirable to use both up-down and down-up transmission,
    fractions. If that is the case, set He3AnalyzerTransmissionFractionMinus directly
    instead of using this helper.
    """
    # The transmission fractions do not share a common time coordinate. Therefore,
    # we cannot concat along a third dimension for the fit, but concat along time,
    # with an additional coordinate for whether polarizations are parallel or
    # antiparallel.
    return sc.concat([updown, downup], 'time').assign_coords(plus_minus=sc.scalar(-1))


def get_he3_transmission_incoming_polarized_from_fit_to_direct_beam(
    plus: He3AnalyzerTransmissionFractionParallel,
    minus: He3AnalyzerTransmissionFractionAntiParallel,
    opacity_function: He3OpacityFunction[Analyzer],
    transmission_empty_glass: He3TransmissionEmptyGlass[Analyzer],
) -> TransmissionFunction[Analyzer]:
    """
    Transmission function for the analyzer, computed with incoming polarized beam.

    This is composed from the opacity-function and the polarization-function.
    The implementation fits a time- and wavelength-dependent equation and returns
    the fitted T(t, lambda).
    """
    if (plus_minus := plus.coords.get('plus_minus')) is not None:
        if not sc.all(plus_minus == sc.scalar(1)):
            raise ValueError('Expected plus-minus coordinate of plus channel to be +1.')
    else:
        plus = plus.assign_coords(plus_minus=sc.scalar(1))
    if (plus_minus := minus.coords.get('plus_minus')) is not None:
        if not sc.all(plus_minus == sc.scalar(-1)):
            raise ValueError(
                'Expected plus-minus coordinate of minus channel to be -1.'
            )
    else:
        minus = minus.assign_coords(plus_minus=sc.scalar(-1))

    transmission_fraction = sc.concat([plus, minus], 'time')

    def expected_transmission(
        wavelength: sc.Variable,
        time: sc.Variable,
        plus_minus: sc.Variable,
        C: sc.Variable,
        T1: sc.Variable,
    ) -> sc.Variable:
        polarization_function = He3PolarizationFunction[PolarizingElement](C=C, T1=T1)
        return He3TransmissionFunction(
            opacity_function=opacity_function,
            polarization_function=polarization_function,
            transmission_empty_glass=transmission_empty_glass,
        )(time=time, wavelength=wavelength, plus_minus=plus_minus)

    popt, _ = sc.curve_fit(
        ['wavelength', 'time', 'plus_minus'],
        expected_transmission,
        _with_midpoints(transmission_fraction, 'wavelength'),
        p0={'C': sc.scalar(0.8, unit=''), 'T1': sc.scalar(400000.0, unit='s')},
    )
    # TODO Consider including variances from fit
    polarization_function = He3PolarizationFunction[PolarizingElement](
        C=sc.values(popt['C']).data, T1=sc.values(popt['T1']).data
    )
    return He3TransmissionFunction[PolarizingElement](
        opacity_function=opacity_function,
        polarization_function=polarization_function,
        transmission_empty_glass=transmission_empty_glass,
    )


def compute_direct_beam(
    data: sc.DataArray,
    q_range: sc.Variable,
    background_q_range: sc.Variable,
) -> sc.DataArray:
    """
    Compute background-subtracted direct beam function.

    The input must be normalized data, not counts.
    """
    if data.bins.unit != '':
        raise ValueError(f'Input data must be normalized, got unit {data.unit}.')
    if q_range.max() > background_q_range.min():
        raise ValueError('Background range must be after direct beam range.')
    if q_range.min() < sc.scalar(0.0, unit='1/angstrom'):
        raise ValueError('Q-range must be positive.')
    q_range = q_range**2
    background_q_range = background_q_range**2
    start_db = q_range[0]
    stop_db = q_range[-1]
    start_bg = background_q_range[0]
    stop_bg = background_q_range[-1]
    # Simple approach for now: Assume we can treat this as rotation invariant
    qx = data.bins.coords['Qx']
    qy = data.bins.coords['Qy']
    data.bins.coords['Q_squared'] = qx**2 + qy**2
    # The input is binned in time and wavelength, we simply take the per-bin mean
    # without changes.
    beam_region = data.bins['Q_squared', start_db:stop_db].bins.mean()
    background = data.bins['Q_squared', start_bg:stop_bg].bins.mean()
    return beam_region - background


DirectBeamQRange = NewType('DirectBeamQRange', sc.Variable)
"""Q-range defining the direct beam region in a direct beam measurement."""

DirectBeamBackgroundQRange = NewType('DirectBeamBackgroundQRange', sc.Variable)
"""Q-range defining the direct beam background region in a direct beam measurement."""

ReducedDirectBeamDataNoCell = NewType('ReducedDirectBeamDataNoCell', sc.DataArray)


class ReducedDirectBeamData(
    sl.ScopeTwoParams[PolarizingElement, PolarizationState, sc.DataArray], sc.DataArray
):
    """Direct beam data for a given cell, as a function of wavelength and time."""


def direct_beam(
    data: ReducedDirectBeamDataNoCell,
    q_range: DirectBeamQRange,
    background_q_range: DirectBeamBackgroundQRange,
) -> DirectBeamNoCell:
    """
    Returns the direct beam function without any cells.

    The result is background-subtracted and returned as function of wavelength.
    Other dimensions of the input are preserved. In particular, the time dimension,
    corresponding to different direct beam measurements, is preserved.
    """
    return DirectBeamNoCell(
        compute_direct_beam(
            data=data,
            q_range=q_range,
            background_q_range=background_q_range,
        )
    )


def direct_beam_with_cell(
    data: ReducedDirectBeamData[PolarizingElement, PolarizationState],
    q_range: DirectBeamQRange,
    background_q_range: DirectBeamBackgroundQRange,
) -> He3DirectBeam[PolarizingElement, PolarizationState]:
    """
    Returns the direct beam function for a given cell.

    The result is background-subtracted and returned as function of wavelength and
    wall-clock time. The time dependence is coarse, i.e., due to different time
    intervals at which the direct beam is measured.
    """
    return He3DirectBeam[PolarizingElement, PolarizationState](
        compute_direct_beam(
            data=data,
            q_range=q_range,
            background_q_range=background_q_range,
        )
    )


providers = (he3_opacity_from_cell_params, direct_beam, direct_beam_with_cell)


def He3CellWorkflow(
    *, in_situ: bool = True, incoming_polarized: bool = False
) -> sl.Pipeline:
    """
    Workflow for computing transmission functions for He3 cells.

    This can handle polarizers as well as analyzers.

    Parameters
    ----------
    in_situ :
        Whether to use an in-situ definition of the cell opacity based on cell
        parameters, or an ex-situ definition based on direct beam data. The latter
        requires a direct-beam measurement with depolarized cells.
    incoming_polarized :
        Whether the incoming beam for computing the analyzer transmission is polarized.
        This is the case in beamlines with a supermirror polarizer, but also if the
        polarizer is not removed from the beam during the analyzer transmission
        measurement.
    """
    workflow = sl.Pipeline(providers)
    if in_situ:
        workflow.insert(he3_opacity_function_from_cell_opacity)
    else:
        workflow.insert(he3_opacity_function_from_beam_data)
    # Note that the incoming-unpolarized function is inserted even if
    # incoming_polarized=True, since the incoming-unpolarized function is still
    # required for computing the *polarizer* transmission calculation.
    workflow.insert(get_he3_transmission_incoming_unpolarized_from_fit_to_direct_beam)
    if incoming_polarized:
        workflow.insert(transmission_fraction_analyzer_parallel)
        workflow.insert(transmission_fraction_analyzer_antiparallel)
        workflow.insert(get_he3_transmission_incoming_polarized_from_fit_to_direct_beam)
    return workflow
